@@ -1,6 +1,6 @@
 /*
     This file is part of darktable,
-    Copyright (C) 2009-2023 darktable developers.
+    Copyright (C) 2009-2025 darktable developers.
 
     darktable is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -52,6 +52,8 @@ typedef struct _dt_job_t
   dt_progress_t *progress;
 
   char description[DT_CONTROL_DESCRIPTION_LEN];
+  dt_view_type_flags_t view_creator;
+  gboolean is_synchronous;
 } _dt_job_t;
 
 /** check if two jobs are to be considered equal. a simple memcmp won't work since the mutexes probably won't
@@ -82,7 +84,7 @@ static void _control_job_set_state(_dt_job_t *job,
   dt_pthread_mutex_lock(&job->state_mutex);
   if(state >= DT_JOB_STATE_FINISHED  && job->state != DT_JOB_STATE_RUNNING && job->progress)
   {
-    dt_control_progress_destroy(darktable.control, job->progress);
+    dt_control_progress_destroy(job->progress);
     job->progress = NULL;
   }
   job->state = state;
@@ -130,7 +132,8 @@ void *dt_control_job_get_params(const _dt_job_t *job)
 dt_job_t *dt_control_job_create(dt_job_execute_callback execute,
                                 const char *msg, ...)
 {
-  _dt_job_t *job = (_dt_job_t *)calloc(1, sizeof(_dt_job_t));
+  if(!dt_control_running()) return NULL;
+  _dt_job_t *job = calloc(1, sizeof(_dt_job_t));
   if(!job) return NULL;
 
   va_list ap;
@@ -140,17 +143,28 @@ dt_job_t *dt_control_job_create(dt_job_execute_callback execute,
 
   job->execute = execute;
   job->state = DT_JOB_STATE_INITIALIZED;
+  job->view_creator = dt_view_get_current();
 
   dt_pthread_mutex_init(&job->state_mutex, NULL);
   dt_pthread_mutex_init(&job->wait_mutex, NULL);
   return job;
 }
 
+dt_view_type_flags_t dt_control_job_get_view_creator(const dt_job_t *job)
+{
+  return job && job->view_creator;
+}
+
+gboolean dt_control_job_is_synchronous(const dt_job_t *job)
+{
+  return job && job->is_synchronous;
+}
+
 void dt_control_job_dispose(_dt_job_t *job)
 {
   if(!job) return;
   if(job->progress)
-    dt_control_progress_destroy(darktable.control, job->progress);
+    dt_control_progress_destroy(job->progress);
   job->progress = NULL;
   _control_job_set_state(job, DT_JOB_STATE_DISPOSED);
   if(job->params_destroy)
@@ -160,19 +174,38 @@ void dt_control_job_dispose(_dt_job_t *job)
   free(job);
 }
 
-void dt_control_job_set_state_callback(_dt_job_t *job, dt_job_state_change_callback cb)
+static inline const char *_queuename(const dt_job_queue_t id)
 {
-  // once the job got added to the queue it may not be changed from the outside
-  if(dt_control_job_get_state(job) != DT_JOB_STATE_INITIALIZED)
-    return; // get_state returns DISPOSED when job == NULL
-  job->state_changed_cb = cb;
+  switch(id)
+  {
+    case DT_JOB_QUEUE_USER_FG:      return "DT_JOB_QUEUE_USER_FG";
+    case DT_JOB_QUEUE_SYSTEM_FG:    return "DT_JOB_QUEUE_SYSTEM_FG";
+    case DT_JOB_QUEUE_USER_BG:      return "DT_JOB_QUEUE_USER_BG";
+    case DT_JOB_QUEUE_USER_EXPORT:  return "DT_JOB_QUEUE_USER_EXPORT";
+    case DT_JOB_QUEUE_SYSTEM_BG:    return "DT_JOB_QUEUE_SYSTEM_BG";
+    case DT_JOB_QUEUE_SYNCHRONOUS:  return "DT_JOB_QUEUE_SYNCHRONOUS";
+    default:                        return "???";
+  }
+}
+// We don't want to log dt_get_wtime() as we already show the stamp
+static void _control_job_print(_dt_job_t *job, const char *info, const char *err, int32_t res)
+{
+  if(!job) return;
+  dt_print(DT_DEBUG_CONTROL, "[%s]\t%02d %s %s | queue: %s | priority: %d",
+    info, res, err, job->description, _queuename(job->queue), job->priority);
 }
 
 
-static void dt_control_job_print(_dt_job_t *job)
+static __thread int32_t threadid = -1;
+// As threadid is `per thread` we don't have to use atomics
+static inline int32_t _control_get_threadid()
 {
-  if(!job) return;
-  dt_print(DT_DEBUG_CONTROL, "%s | queue: %d | priority: %d", job->description, job->queue, job->priority);
+  return threadid > -1 ? threadid : darktable.control->num_threads;
+}
+
+static inline int32_t _control_get_threadid_res()
+{
+  return threadid > -1 ? threadid : DT_CTL_WORKER_RESERVED;
 }
 
 void dt_control_job_cancel(_dt_job_t *job)
@@ -180,25 +213,6 @@ void dt_control_job_cancel(_dt_job_t *job)
   _control_job_set_state(job, DT_JOB_STATE_CANCELLED);
 }
 
-void dt_control_job_wait(_dt_job_t *job)
-{
-  if(!job) return;
-  dt_job_state_t state = dt_control_job_get_state(job);
-
-  // NOTE: could also use signals.
-
-  /* if job execution is not finished let's wait for it */
-  if(state == DT_JOB_STATE_RUNNING || state == DT_JOB_STATE_CANCELLED)
-  {
-    // once the job finishes, it unlocks the mutex
-    // so by locking the mutex here, we will only get the lock once the job
-    // has finished and unlocked it.
-    dt_pthread_mutex_lock(&job->wait_mutex);
-    // yay, the job finished, we got the lock. nothing more to do.
-    dt_pthread_mutex_unlock(&job->wait_mutex);
-  }
-}
-
 static gboolean _control_run_job_res(dt_control_t *control, int32_t res)
 {
   if(((unsigned int)res) >= DT_CTL_WORKER_RESERVED)
@@ -219,9 +233,7 @@ static gboolean _control_run_job_res(dt_control_t *control, int32_t res)
   dt_pthread_mutex_lock(&job->wait_mutex);
   if(dt_control_job_get_state(job) == DT_JOB_STATE_QUEUED)
   {
-    dt_print(DT_DEBUG_CONTROL, "[run_job+] %02d %f ", res, dt_get_wtime());
-    dt_control_job_print(job);
-    dt_print(DT_DEBUG_CONTROL, "\n");
+    _control_job_print(job, "run_job+", "", res);
 
     _control_job_set_state(job, DT_JOB_STATE_RUNNING);
 
@@ -229,9 +241,7 @@ static gboolean _control_run_job_res(dt_control_t *control, int32_t res)
     job->result = job->execute(job);
 
     _control_job_set_state(job, DT_JOB_STATE_FINISHED);
-    dt_print(DT_DEBUG_CONTROL, "[run_job-] %02d %f ", res, dt_get_wtime());
-    dt_control_job_print(job);
-    dt_print(DT_DEBUG_CONTROL, "\n");
+    _control_job_print(job, "run_job-", "", res);
   }
   dt_pthread_mutex_unlock(&job->wait_mutex);
   dt_control_job_dispose(job);
@@ -287,7 +297,7 @@ static _dt_job_t *_control_schedule_job(dt_control_t *control)
   if(winner_queue == DT_JOB_QUEUE_USER_EXPORT) control->export_scheduled = TRUE;
 
   // and place it in scheduled job array (for job deduping)
-  control->job[dt_control_get_threadid()] = job;
+  control->job[_control_get_threadid()] = job;
 
   // increment the priorities of the others
   for(int i = 0; i < DT_JOB_QUEUE_MAX; i++)
@@ -303,10 +313,7 @@ static _dt_job_t *_control_schedule_job(dt_control_t *control)
 
 static void _control_job_execute(_dt_job_t *job)
 {
-  dt_print(DT_DEBUG_CONTROL, "[run_job+] %02d %f ", DT_CTL_WORKER_RESERVED + dt_control_get_threadid(),
-           dt_get_wtime());
-  dt_control_job_print(job);
-  dt_print_nts(DT_DEBUG_CONTROL, "\n");
+  _control_job_print(job, "run_job+", "", DT_CTL_WORKER_RESERVED + _control_get_threadid());
 
   _control_job_set_state(job, DT_JOB_STATE_RUNNING);
 
@@ -314,11 +321,7 @@ static void _control_job_execute(_dt_job_t *job)
   job->result = job->execute(job);
 
   _control_job_set_state(job, DT_JOB_STATE_FINISHED);
-
-  dt_print(DT_DEBUG_CONTROL, "[run_job-] %02d %f ", DT_CTL_WORKER_RESERVED + dt_control_get_threadid(),
-           dt_get_wtime());
-  dt_control_job_print(job);
-  dt_print_nts(DT_DEBUG_CONTROL, "\n");
+  _control_job_print(job, "run_job-", "", DT_CTL_WORKER_RESERVED + _control_get_threadid());
 }
 
 static gboolean _control_run_job(dt_control_t *control)
@@ -336,20 +339,20 @@ static gboolean _control_run_job(dt_control_t *control)
 
   // remove the job from scheduled job array (for job deduping)
   dt_pthread_mutex_lock(&control->queue_mutex);
-  control->job[dt_control_get_threadid()] = NULL;
+  control->job[_control_get_threadid()] = NULL;
   if(job->queue == DT_JOB_QUEUE_USER_EXPORT) control->export_scheduled = FALSE;
   dt_pthread_mutex_unlock(&control->queue_mutex);
 
   // and free it
   dt_control_job_dispose(job);
+  dt_atomic_sub_int(&control->pending_jobs, 1);
 
   return FALSE;
 }
 
-gboolean dt_control_add_job_res(dt_control_t *control,
-                                _dt_job_t *job,
-                                int32_t res)
+gboolean dt_control_add_job_res(_dt_job_t *job, const int32_t res)
 {
+  dt_control_t *control = darktable.control;
   if(((unsigned int)res) >= DT_CTL_WORKER_RESERVED || !job)
   {
     dt_control_job_dispose(job);
@@ -366,9 +369,7 @@ gboolean dt_control_add_job_res(dt_control_t *control,
     dt_control_job_dispose(control->job_res[res]);
   }
 
-  dt_print(DT_DEBUG_CONTROL, "[add_job_res] %d | ", res);
-  dt_control_job_print(job);
-  dt_print_nts(DT_DEBUG_CONTROL, "\n");
+  _control_job_print(job, "add_job_res", "", res);
 
   _control_job_set_state(job, DT_JOB_STATE_QUEUED);
   control->job_res[res] = job;
@@ -383,20 +384,20 @@ gboolean dt_control_add_job_res(dt_control_t *control,
   return FALSE;
 }
 
-gboolean dt_control_add_job(dt_control_t *control,
-                            dt_job_queue_t queue_id,
-                            _dt_job_t *job)
+gboolean dt_control_add_job(dt_job_queue_t queue_id, _dt_job_t *job)
 {
-  if(((unsigned int)queue_id) >= DT_JOB_QUEUE_MAX || !job)
+  dt_control_t *control = darktable.control;
+  if((((unsigned int)queue_id) >= DT_JOB_QUEUE_MAX && queue_id != DT_JOB_QUEUE_SYNCHRONOUS) || !job)
   {
     dt_control_job_dispose(job);
     return TRUE;
   }
 
-  if(!control->running)
+  if(!dt_control_running() || queue_id == DT_JOB_QUEUE_SYNCHRONOUS)
   {
     // whatever we are adding here won't be scheduled as the system isn't running. execute it synchronous instead.
     dt_pthread_mutex_lock(&job->wait_mutex); // is that even needed?
+    job->is_synchronous = TRUE;
     _control_job_execute(job);
     dt_pthread_mutex_unlock(&job->wait_mutex);
 
@@ -413,10 +414,9 @@ gboolean dt_control_add_job(dt_control_t *control,
   GList **queue = &control->queues[queue_id];
   size_t length = control->queue_length[queue_id];
 
-  dt_print(DT_DEBUG_CONTROL, "[add_job] %zu | ", length);
-  dt_control_job_print(job);
-  dt_print_nts(DT_DEBUG_CONTROL, "\n");
+  _control_job_print(job, "add_job", "", (int32_t)length);
 
+  dt_atomic_add_int(&control->pending_jobs, 1);
   if(queue_id == DT_JOB_QUEUE_SYSTEM_FG)
   {
     // this is a stack with limited size and bubble up and all that stuff
@@ -425,17 +425,16 @@ gboolean dt_control_add_job(dt_control_t *control,
     // check if we have already scheduled the job
     for(int k = 0; k < control->num_threads; k++)
     {
-      _dt_job_t *other_job = (_dt_job_t *)control->job[k];
+      _dt_job_t *other_job = control->job[k];
       if(_control_job_equal(job, other_job))
       {
-        dt_print(DT_DEBUG_CONTROL, "[add_job] found job already in scheduled: ");
-        dt_control_job_print(other_job);
-        dt_print_nts(DT_DEBUG_CONTROL, "\n");
+        _control_job_print(other_job, "add_job", "found job already in scheduled:", -1);
 
         dt_pthread_mutex_unlock(&control->queue_mutex);
 
         _control_job_set_state(job, DT_JOB_STATE_DISCARDED);
         dt_control_job_dispose(job);
+        dt_atomic_sub_int(&control->pending_jobs, 1);
 
         return 0; // there can't be any further copy
       }
@@ -444,15 +443,14 @@ gboolean dt_control_add_job(dt_control_t *control,
     // if the job is already in the queue -> move it to the top
     for(GList *iter = *queue; iter; iter = g_list_next(iter))
     {
-      _dt_job_t *other_job = (_dt_job_t *)iter->data;
+      _dt_job_t *other_job = iter->data;
       if(_control_job_equal(job, other_job))
       {
-        dt_print(DT_DEBUG_CONTROL, "[add_job] found job already in queue: ");
-        dt_control_job_print(other_job);
-        dt_print_nts(DT_DEBUG_CONTROL, "\n");
+        _control_job_print(other_job, "add_job", "found job already in queue", -1);
 
         *queue = g_list_delete_link(*queue, iter);
         length--;
+        dt_atomic_sub_int(&control->pending_jobs, 1);
 
         job_for_disposal = job;
 
@@ -469,10 +467,11 @@ gboolean dt_control_add_job(dt_control_t *control,
     if(length > DT_CONTROL_MAX_JOBS)
     {
       GList *last = g_list_last(*queue);
-      _control_job_set_state((_dt_job_t *)last->data, DT_JOB_STATE_DISCARDED);
-      dt_control_job_dispose((_dt_job_t *)last->data);
+      _control_job_set_state(last->data, DT_JOB_STATE_DISCARDED);
+      dt_control_job_dispose(last->data);
       *queue = g_list_delete_link(*queue, last);
       length--;
+      dt_atomic_sub_int(&control->pending_jobs, 1);
     }
 
     control->queue_length[queue_id] = length;
@@ -504,20 +503,6 @@ gboolean dt_control_add_job(dt_control_t *control,
   return FALSE;
 }
 
-static __thread int threadid = -1;
-
-int32_t dt_control_get_threadid()
-{
-  if(threadid > -1) return threadid;
-  return darktable.control->num_threads;
-}
-
-static inline int32_t _control_get_threadid_res()
-{
-  if(threadid > -1) return threadid;
-  return DT_CTL_WORKER_RESERVED;
-}
-
 static void *_control_work_res(void *ptr)
 {
 #ifdef _OPENMP // need to do this in every thread
@@ -530,10 +515,10 @@ static void *_control_work_res(void *ptr)
   snprintf(name, sizeof(name), "worker res %d", threadid);
   dt_pthread_setname(name);
   free(params);
-  int32_t threadid_res = _control_get_threadid_res();
+  const int32_t threadid_res = _control_get_threadid_res();
   while(dt_control_running())
   {
-    // dt_print(DT_DEBUG_CONTROL, "[control_work] %d\n", threadid_res);
+    // dt_print(DT_DEBUG_CONTROL, "[control_work] %d", threadid_res);
     if(_control_run_job_res(s, threadid_res))
     {
       // wait for a new job.
@@ -575,10 +560,9 @@ static void *_control_work(void *ptr)
   snprintf(name, sizeof(name), "worker %d", threadid);
   dt_pthread_setname(name);
   free(params);
-  // int32_t threadid = dt_control_get_threadid();
+
   while(dt_control_running())
   {
-    // dt_print(DT_DEBUG_CONTROL, "[control_work] %d\n", threadid);
     if(_control_run_job(control))
     {
       // wait for a new job.
@@ -592,27 +576,27 @@ static void *_control_work(void *ptr)
 
 // convenience functions to have a progress bar for the job.
 // this allows to show the gui indicator of the job even before it got scheduled
-void dt_control_job_add_progress(dt_job_t *job, const char *message, gboolean cancellable)
+void dt_control_job_add_progress(dt_job_t *job, const char *message, const gboolean cancellable)
 {
   if(!job) return;
-  job->progress = dt_control_progress_create(darktable.control, TRUE, message);
+  job->progress = dt_control_progress_create(TRUE, message);
   if(cancellable)
-    dt_control_progress_attach_job(darktable.control, job->progress, job);
+    dt_control_progress_attach_job(job->progress, job);
 }
 
 void dt_control_job_set_progress_message(dt_job_t *job, const char *message)
 {
   if(!job || !job->progress) return;
-  dt_control_progress_set_message(darktable.control, job->progress, message);
+  dt_control_progress_set_message(job->progress, message);
 }
 
-void dt_control_job_set_progress(dt_job_t *job, double value)
+void dt_control_job_set_progress(dt_job_t *job, const double value)
 {
   if(!job || !job->progress) return;
-  dt_control_progress_set_progress(darktable.control, job->progress, value);
+  dt_control_progress_set_progress(job->progress, value);
 }
 
-double dt_control_job_get_progress(dt_job_t *job)
+double dt_control_job_get_progress(const dt_job_t *job)
 {
   if(!job || !job->progress) return -1.0;
   return dt_control_progress_get_progress(job->progress);
@@ -620,52 +604,67 @@ double dt_control_job_get_progress(dt_job_t *job)
 
 
 // moved out of control.c to be able to make some helper functions static
-void dt_control_jobs_init(dt_control_t *control)
+void dt_control_jobs_init()
 {
+  dt_control_t *control = darktable.control;
   // start threads
   control->num_threads = dt_worker_threads();
   control->thread = (pthread_t *)calloc(control->num_threads, sizeof(pthread_t));
   control->job = (dt_job_t **)calloc(control->num_threads, sizeof(dt_job_t *));
-  dt_pthread_mutex_lock(&control->run_mutex);
-  control->running = TRUE;
-  dt_pthread_mutex_unlock(&control->run_mutex);
+
+  g_atomic_int_set(&control->running, DT_CONTROL_STATE_RUNNING);
+
+  int err = 0; // collected errors while creating all threads
+
   for(int k = 0; k < control->num_threads; k++)
   {
-    worker_thread_parameters_t *params
-        = (worker_thread_parameters_t *)calloc(1, sizeof(worker_thread_parameters_t));
+    worker_thread_parameters_t *params = calloc(1, sizeof(worker_thread_parameters_t));
     params->self = control;
     params->threadid = k;
-    dt_pthread_create(&control->thread[k], _control_work, params);
+    err |= dt_pthread_create(&control->thread[k], _control_work, params);
   }
 
   /* create queue kicker thread */
-  dt_pthread_create(&control->kick_on_workers_thread, _control_worker_kicker, control);
+  err |= dt_pthread_create(&control->kick_on_workers_thread, _control_worker_kicker, control);
 
   for(int k = 0; k < DT_CTL_WORKER_RESERVED; k++)
   {
     control->job_res[k] = NULL;
     control->new_res[k] = 0;
-    worker_thread_parameters_t *params
-        = (worker_thread_parameters_t *)calloc(1, sizeof(worker_thread_parameters_t));
+    worker_thread_parameters_t *params = calloc(1, sizeof(worker_thread_parameters_t));
     params->self = control;
     params->threadid = k;
-    dt_pthread_create(&control->thread_res[k], _control_work_res, params);
+    err |= dt_pthread_create(&control->thread_res[k], _control_work_res, params);
   }
   /* create thread taking care of connecting gphoto2 devices */
 #ifdef HAVE_GPHOTO2
-  dt_pthread_create(&control->update_gphoto_thread, dt_update_cameras_thread, control);
+  err |= dt_pthread_create(&control->update_gphoto_thread, dt_update_cameras_thread, control);
 #endif
+  if(err != 0)
+    dt_print(DT_DEBUG_ALWAYS, "[dt_control_jobs_init] couldn't create all threads, problems ahead");
 }
 
-void dt_control_jobs_cleanup(dt_control_t *control)
+void dt_control_jobs_cleanup()
 {
+  dt_control_t *control = darktable.control;
   free(control->job);
+  control->job = NULL;
   free(control->thread);
+  control->thread = NULL;
 }
 
+int dt_control_jobs_pending()
+{
+  dt_control_t *control = darktable.control;
+  if(!control) return 0;
+
+  int pending = dt_atomic_get_int(&control->pending_jobs);
+  if(darktable.backthumbs.running)
+    pending--;
+  return pending;
+}
 // clang-format off
 // modelines: These editor modelines have been set for all relevant files by tools/update_modelines.py
 // vim: shiftwidth=2 expandtab tabstop=2 cindent
 // kate: tab-indents: off; indent-width 2; replace-tabs on; indent-mode cstyle; remove-trailing-spaces modified;
 // clang-format on
-
